;######################################################
;##                    IMAGES                        ##
;##--------------------------------------------------##
;## Iamges are represented internally by carrrays    ##
;## of ctype _float and with dimension 2.            ##
;## To ensure an abstraction layer above the carray  ##
;## we add meaningful accessors (e.g. martix-width   ##
;## or image-height for the dimensions accessors.    ##
;## In addition mapping and reduce functions are     ##
;## available both in-place and functional.          ##
;######################################################


; create an image from width/height
; and optional: initial color
(define (make-image width height . init-val)
  (if (pair? init-val)
      (make-carray _float (list width height)(car init-val) )
      (make-carray _float (list width height) )))

; image accessor: data (cvector) of an image
(define (image-data image)
  (carray-data image))

; image accessor: width of an image
(define (image-width image)
  (car (carray-dimensions image)))

; image accessor: height of an image
(define (image-height image)
  (cadr (carray-dimensions image)))

; better accessors - ref
(define (image-ref image x y)
  (cvector-ref (image-data image) (+ x (* (image-width image) y ))))
  
; better accessors - set
(define (image-set! image x y val)
  (cvector-set! (image-data image) (+ x (* (image-width image) y )) val))


; copying of images
(define (copy-image image)
  (copy-carray image))

; image -> list of lists conversion
(define (image->list image)
  (carray->list image))

; list -> image conversion 
(define (list->image lst)
  (list->carray lst))

; map a function (lambda (pixel_intensity 1 .. 2 .. 3 .. n) -> a_new_intensity)
; same as above but in-place!
; Changes the first given image in the list!
(define (image-map func . images)
  (if (null? images)
      '()
      (let* ((my_image (copy-image (car images)))
             (width (image-width my_image))
             (height (image-height my_image))
             (image_vec (image-data my_image))
             (map_func (lambda(image i) 
                         (cvector-ref (image-data image) i))))
    (do ((i 0 (+ i 1)))
      ((= i (* width height)) my_image)
      (cvector-set! image_vec i  (apply func (map (lambda(image) (map_func image i)) images)))))))
  

; same as above but in-place!
; Changes the first given image in the list!
(define (image-map! func . images)
  (if (null? images)
      '()
      (let* ((my_image (car images))
             (width (image-width my_image))
             (height (image-height my_image))
             (image_vec (image-data my_image))
             (map_func (lambda(image i) 
                         (cvector-ref (image-data image) i))))
    (do ((i 0 (+ i 1)))
      ((= i (* width height)) my_image)
      (cvector-set! image_vec i  (apply func (map (lambda(image) (map_func image i)) images)))))))
  
; reduce for images
(define (image-reduce func image seed)
  (let* ((width (image-width image))
         (height (image-height image))
         (image_vec (image-data image))
         (reduce_var seed))
    (do ((i 0 (+ i 1)))
      ((= i (* width height)) reduce_var)
      (set! reduce_var  (func reduce_var (cvector-ref image_vec i))))))
  
; apply a function (lambda (x y) -> void)
; to an image in-place!
(define (image-for-each-index func image)
  (let* ((width (image-width image))
         (height (image-height image)))
    (do ((y 0 (+ y 1)))
      ((= y height)  image)
      (do ((x 0 (+ x 1)))
        ((= x width) )
        (func x y)))))

  
  
;######################################################
;##                    MATRICES                      ##
;##--------------------------------------------------##
;## Matrices are represented internally by carrrays  ##
;## of ctype _double and with dimension 2.           ##
;## To ensure an abstraction layer above the carray  ##
;## we add meaningful accessors (e.g. martix-cols    ##
;## or matrix-rows for the dimensions accessors.     ##
;## In addition a  matrix multiplication is avail.   ##
;######################################################


; creation of an image from width/height 
; and optional: initial value
(define (make-matrix rows cols . init-val)
  (if (pair? init-val)
      (make-carray _double (list rows cols)(car init-val) )
      (make-carray _double (list rows cols) )))
  
; matrix-accessors - rows of matrix
(define (matrix-rows mat)
  (car (carray-dimensions mat)))

; matrix-accessors - columns of matrix
(define (matrix-cols mat)
  (cadr (carray-dimensions mat)))
  
; matrix-accessors - data inside matrix
(define (matrix-data matrix)
  (carray-data matrix))
  
; better accessors - ref
(define (matrix-ref mat r c)
  (cvector-ref (carray-data mat) (+ r (* (matrix-rows mat) c ))))
  
; better accessors - set
(define (matrix-set! mat r c val)
  (cvector-set! (carray-data mat) (+ r (* (matrix-rows mat) c )) val))
 
; copying of matrices
(define (copy-matrix mat)
  (copy-carray mat))
  
; matrix -> list of lists conversion
(define (matrix->list mat)
  (carray->list mat))

; list -> matrix conversion
(define (list->matrix lst)
  (list->carray lst))

; multiplication of two matrices 
(define (matrix-mult mat1 mat2)
  (if (and (= (matrix-rows mat1)
              (matrix-cols mat2))
           (= (matrix-cols mat1)
              (matrix-rows mat2)))
        (let ((new-matrix (make-matrix (matrix-rows mat2) (matrix-cols mat1) 0.0)))
          (do ((c 0 (+ c 1)))
            ((= c (matrix-cols mat1)) new-matrix)
            (do ((r 0 (+ r 1)))
              ((= r  (matrix-rows mat2)))
              (do ((i 0 (+ i 1)))
                ((= i (matrix-rows mat1)))
                (matrix-set! new-matrix r c (+ (matrix-ref new-matrix r c)
                                               (* (matrix-ref mat1 i c)
                                                  (matrix-ref mat2 r i))))))))
        (error "matrix sizes do not match")))